from unittest import mock

from nxdomain import (
    read_block_list,
    BlockList,
    BlockListType,
    download_and_generate,
    DnsmasqGenerator,
    BindGenerator,
)


def test_read_simple(testdir):
    f = BlockList(str(testdir / "simple.txt"), BlockListType.simple)
    bl = list(read_block_list(f))

    assert len(bl) == 3
    assert set(bl) == {"example.org", "example.com", "example.net"}


def test_read_hosts(testdir):
    f = BlockList(str(testdir / "hosts.txt"), BlockListType.hosts)
    bl = list(read_block_list(f))

    assert len(bl) == 4
    assert set(bl) == {"example.org", "example.com"}


def test_read_http(testdir):
    with mock.patch("urllib.request.urlopen") as urlopen:
        urlopen.return_value = (testdir / "simple.txt").open("rb")
        f = BlockList("http://example.org/simple.txt", BlockListType.simple)
        bl = list(read_block_list(f))

        assert len(bl) == 3
        assert set(bl) == {"example.org", "example.com", "example.net"}


def test_generate_dnsmasq(testdir, tmp_path):
    bls = [BlockList(str(testdir / "simple.txt"), BlockListType.simple)]
    generator = DnsmasqGenerator()
    out_path = tmp_path / "out.conf"
    download_and_generate(bls, generator, str(out_path))

    assert out_path.open().read() == "\n".join(
        [
            "address=/example.com/#",
            "address=/example.net/#",
            "address=/example.org/#",
            "",
        ]
    )


def test_generate_bind(testdir, tmp_path):
    bls = [BlockList(str(testdir / "simple.txt"), BlockListType.simple)]
    generator = BindGenerator()
    out_path = tmp_path / "out.zone"
    download_and_generate(bls, generator, str(out_path))

    assert (
        out_path.open().read()
        == """@ 3600 IN SOA @ hostmaster 0 86400 7200 2592000 86400
@ 3600 IN NS LOCALHOST.
example.com 3600 IN CNAME .
example.net 3600 IN CNAME .
example.org 3600 IN CNAME .
"""
    )

    download_and_generate(bls, generator, str(out_path))
    assert "hostmaster 1" in out_path.open().readline()



import abc
import enum
import itertools
import logging
import re
import urllib.request
from typing import Iterable, BinaryIO, NamedTuple, List

import dns.zone  # type: ignore

# Certainly not spec-compliant, but good enough for sanity checks.
DOMAIN_NAME_REGEXP = re.compile(
    r"\b((?:(?=[a-z0-9-]{1,63}\.)(xn--)?[a-z0-9]+(-[a-z0-9]+)*\.)+[a-z]{1,63})\b"
)
HOSTS_REGEXP = re.compile(r"^\S+\s+(.+)$")


class BlockListType(enum.Enum):
    simple = "simple"
    hosts = "hosts"


class BlockList(NamedTuple):
    uri: str
    type: BlockListType


class BlockListGenerator(abc.ABC):
    def generate(self, domains: Iterable[str], filename: str):
        ...  # pragma: no cover


class DnsmasqGenerator(BlockListGenerator):
    def generate(self, domains: Iterable[str], filename: str):
        domains = sorted(set(domains))

        with open(filename, "w") as f:
            for domain in domains:
                f.write(f"address=/{domain}/#\n")

        logging.info("block list has %d domains", len(domains))


class BindGenerator(BlockListGenerator):
    def generate(self, domains: Iterable[str], filename: str):
        ttl = 3600
        origin = "rpz.local"

        def read_header():
            with open(filename, "r") as f:
                for line in f:
                    if not line.startswith("@"):
                        break
                    yield line

        try:
            # Read "header" only since it's expensive to parse everything.
            zone = dns.zone.from_text("\n".join(read_header()), origin=origin)
        except (dns.exception.DNSException, IOError):
            zone = dns.zone.Zone(origin)

        existing_soa_set = zone.get_rdataset("@", "SOA")
        zone.nodes.clear()

        soa_set = zone.find_rdataset("@", "SOA", create=True)
        if existing_soa_set is None:
            soa_set.ttl = 3600
            soa_set.add(
                dns.rdata.from_text(
                    "IN", "SOA", "@ hostmaster 0 86400 7200 2592000 86400"
                )
            )
            logging.info("creating zone from scratch")
        else:
            soa_set.update(existing_soa_set)
            # Increment serial.
            serial = soa_set[0].serial + 1
            soa_set.add(soa_set[0].replace(serial=serial))
            logging.info("parsed existing header, new serial is %d", serial)

        in_ns = zone.find_rdataset("@", "NS", create=True)
        in_ns.ttl = ttl
        in_ns.add(dns.rdata.from_text("IN", "NS", "LOCALHOST."))

        in_cname_dot = dns.rdataset.from_text("IN", "CNAME", ttl, ".")
        for domain in domains:
            zone.find_node(domain, create=True).replace_rdataset(in_cname_dot)

        zone.to_file(filename, sorted=True)
        logging.info("block list has %d domains", len(zone.nodes) - 1)


def parse_block_list(list_type: BlockListType, f: BinaryIO) -> Iterable[str]:
    def reader():
        for raw_line in f:
            line = raw_line.decode().strip()
            if not line:
                continue
            if list_type is BlockListType.hosts:
                if (match := HOSTS_REGEXP.match(line)) is not None and (
                    domain := match.group(1).strip()
                ):
                    yield domain
            elif list_type is BlockListType.simple:
                if not line.startswith("#"):
                    yield line
            else:  # pragma: no cover
                raise ValueError(f"unsupported list type {list_type}")

    for line in reader():
        if (m := DOMAIN_NAME_REGEXP.search(line)) is not None and (
            domain := m.group(1).strip()
        ):
            try:
                # Validate syntax and yield.
                yield dns.name.from_text(domain).to_text(omit_final_dot=True)
            except dns.exception.DNSException:  # pragma: no cover
                pass


def read_block_list(block_list: BlockList) -> Iterable[str]:
    if block_list.uri.startswith("http"):
        with urllib.request.urlopen(block_list.uri, timeout=10) as r:
            yield from parse_block_list(block_list.type, r)
            return

    with open(block_list.uri, "rb") as f:
        yield from parse_block_list(block_list.type, f)
        return


def download_and_generate(
    lists: List[BlockList], generator: BlockListGenerator, filename: str
):
    domains = itertools.chain(*(read_block_list(block_list) for block_list in lists))
    generator.generate(domains, filename)



import argparse
import logging
import sys

from nxdomain import (
    BindGenerator,
    BlockList,
    BlockListGenerator,
    BlockListType,
    DnsmasqGenerator,
    download_and_generate,
)


def main_argv(argv):
    p = argparse.ArgumentParser(description="nxdomain -- domain block list management")
    p.add_argument("--out", required=True, help="output zone filename")
    p.add_argument(
        "--format",
        required=True,
        choices=["bind", "dnsmasq"],
        help="output format, BIND zone file or dnsmasq config",
    )
    p.add_argument("--simple", action="append", help="adds a simple block list URI")
    p.add_argument(
        "--hosts",
        action="append",
        help="adds a 'hosts' syntax block list URI",
    )
    args = p.parse_args(argv)

    simple_lists = [BlockList(uri, BlockListType.simple) for uri in (args.simple or [])]
    hosts_lists = [BlockList(uri, BlockListType.hosts) for uri in (args.hosts or [])]
    all_lists = simple_lists + hosts_lists

    if not all_lists:
        p.error("at least one --simple or --hosts is required")
        # Exits.

    generator: BlockListGenerator
    if args.format == "bind":
        generator = BindGenerator()
    elif args.format == "dnsmasq":
        generator = DnsmasqGenerator()

    logging.basicConfig(level=logging.INFO)
    download_and_generate(all_lists, generator, args.out)


def main():
    main_argv(sys.argv[1:])


if __name__ == "__main__":  # pragma: no cover
    main()



from setuptools import setup, find_packages
import os.path

pwd = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(pwd, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="nxdomain",
    version="1.0.1",
    description="Downloads and converts domain block lists for use with BIND/named RPZ files or dnsmasq host files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/zopieux/nxdomain",
    author="Alexandre Macabies",
    author_email="web@zopieux.com",
    install_requires=["dnspython~=2.0"],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'nxdomain=nxdomain.__main__:main'
        ]
    },
    project_urls={
        "Bug Reports": "https://github.com/zopieux/nxdomain/issues",
        "Source": "https://github.com/zopieux/nxdomain",
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    zip_safe=True,
    include_package_data=True,
)
